#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2021 Ina (David Doukhan & Zohra Rezgui- http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from typing import NamedTuple
import dlib
import numpy as np
from .rect import Rect
from .opencv_utils import disp_frame_shapes


class TrackDetection(NamedTuple):
    """
    Atomic element returned by face tracking and detection classes
    """
    #: bounding box
    bbox : Rect
    #: detected face numerical identifier
    face_id : int
    #: face detection confidence
    detect_conf : float
    #: tracking confidence
    track_conf : float

def _matrix_argmax(m):
    x = np.argmax(m)
    dim = m.shape[1]
    return x // dim, x % dim

class Tracker:
    def __init__(self, frame, bb, detect_conf):
        self.t = dlib.correlation_tracker()
        if not isinstance(bb, dlib.drectangle):
            bb = dlib.drectangle(*bb)
        self.t.start_track(frame, bb)
        self.fshape = frame.shape
        self.detect_conf = detect_conf
        self.track_conf = None

    def update(self, frame, verbose=False):
        update_val = self.t.update(frame)

        fh, fw, _ = self.fshape

        e = self.t.get_position()
        x1, y1, x2, y2 = pos = Rect.from_dlib(e)


        if verbose:
            print('update', pos, update_val)
            disp_frame_shapes(frame, [pos])

        if not ((x2 > 0) and (x1 < fw) and (y2 > 0) and (y1 < fh)):
            update_val = -1

        self.detect_conf = None
        self.track_conf = update_val
        return update_val

    def update_from_detection(self, frame, dtc, verbose=False):
        update_val = self.t.update(frame, dtc.bbox.to_dlibFloat())
        if verbose:
            e = self.t.get_position()
            print('dest', dtc.bbox, 'new position', e)
            disp_frame_shapes(frame, [Rect.from_dlib(e), dtc.bbox])
        self.t.start_track(frame, dtc.bbox.to_dlibFloat())
        self.track_conf = update_val
        self.detect_conf = dtc.detect_conf
        return update_val


class TrackerDetector:
    # confidence is estimated as the peak to side-lobe ration returned
    # on tracker's update
    min_confidence = 7

    # output labels
    output_type = TrackDetection
    #out_names = ['bb', 'face_id', 'face_detect_conf', 'tracking_conf']

    def __init__(self, detector, detection_period):
        # dictionnary of tracked objects
        self.d = {}
        # number of instantiated trackers
        # used to provide a numeric identifier to each new Tracker instance
        self.nb_tracker = 0
        # a face/object detection instance
        # could be any face detector class defined in face_detector.py
        self.detector = detector
        # Detection will be performed once every 'detection_period' frames
        # ex: if detection_period is set to 5, detection will be performed
        # for 1 frame, and tracking for the 4 remaining frames
        # if detection_period is set to 1, the detection will be performed on
        # every frame, and tracking will also be performed to know the detected
        # faces belong to the same person
        self.detection_period = detection_period
        # count the amount of processed frames
        # used to switch between detection and tracking every detection_period
        self.iframe = 0


    def update_trackers(self, frame, verbose = False):
        if verbose:
            print('update trackers')
        # if tracked element is lost, remove tracker
        for fid in list(self.d):
            if self.d[fid].update(frame, verbose) < self.min_confidence:
                del self.d[fid]
                if verbose:
                    print('deleting tracker', fid)

    def update_from_detection(self, frame, ldetections, verbose):

        if verbose:
            print('update from detection')

        lkeys = list(self.d.keys())

        # compute intersection over union matrix[#tracker, #detected bounding box]
        # between tracker positions and detected bounding boxes
        ioumat = np.ones((len(lkeys), len(ldetections))) * -1
        for i, k in enumerate(lkeys):
            trackpos = Rect.from_dlib(self.d[k].t.get_position())
            for j, detection in enumerate(ldetections):
                ioumat[i, j] = detection.bbox.iou(trackpos)

        # while matrix not empty and IOU > 70%
        while np.prod(ioumat.shape):
            # find the largest intersection over union
            itracker, idetection = am = _matrix_argmax(ioumat)
            if ioumat[am] <= 0.7:
                break
            # update closest bounding box and trackers and remove them from matrix
            track_score = self.d[lkeys[itracker]].update_from_detection(frame, ldetections[idetection], verbose)
            ioumat = np.delete(ioumat, itracker, axis = 0)
            k = lkeys.pop(itracker)

            # if close bounding box and tracker do not match, delete tracker
            if track_score < self.min_confidence:
                del self.d[k]
            else:
                # if bounding box and detected face match, remove
                # the detection from the set
                ldetections.pop(idetection)
                ioumat = np.delete(ioumat, idetection, axis = 1)


        # remove trackers that do not match any detected box
        for k in lkeys:
            del self.d[k]

        # add new trackers corresponding to detected faces that did not match
        # any existing tracker
        for dtc in ldetections:
            self.d[self.nb_tracker] = Tracker(frame, dtc.bbox, dtc.detect_conf)
            self.nb_tracker += 1


    def __call__(self, frame, verbose=False):

        if self.iframe % self.detection_period == 0:
            self.update_from_detection(frame, self.detector(frame, verbose), verbose)
        else:
            self.update_trackers(frame, verbose)

        lret = []
        for faceid in self.d:
            t = self.d[faceid]
            bb = t.t.get_position()
            lret.append(TrackDetection(Rect.from_dlib(bb), faceid, t.detect_conf, t.track_conf))

        self.iframe += 1

        return lret



#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2022 Ina (David Doukhan - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module :mod:`inaFaceAnalyzer.notebook_utils` contain simple functions allowing to display video in jupyter
and google collab's notebooks. These functions can be used only in a notebook environment
"""


try:
    # To be used only in ipython/collab environments
    from IPython.core.display import display
    from IPython.display import HTML
    from base64 import b64encode

    def notebook_display_remote_vid(video_path, width=600):
        """
        Display a remote video in a jupyter notebook
        Can be used only in a jupyter environment

        Args:
            video_path (str): path or url to the remote video to be displayed
            width (int, optional): video width in the notebook. Defaults to 600 pixels.
        """
        data = '<div align="middle"> <video width=%d controls> <source src=%s> </video></div>' % (width, video_path)
        display(HTML(data))

    def notebook_display_local_vid(video_path, width = 600):
        """
        Display a local video in a Jupyter Notebook
        Compatible with google collab's notebooks

        Args:
            video_path (str): path to the local video.
            width (int, optional): video width in the notebook. Defaults to 600 pixels.

        """

        # thanks https://androidkt.com/how-to-capture-and-play-video-in-google-colab/
        video_file = open(video_path, "r+b").read()
        video_url = f"data:video/mp4;base64,{b64encode(video_file).decode()}"
        return HTML(f"""<video width={width} controls><source src="{video_url}"></video>""")

except:
    pass



#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2021 Ina (David Doukhan - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from typing import NamedTuple
import dlib

class Rect(NamedTuple):
    """
    This class is an internal data structure allowing to manipulate rectangle shapes
    """

    #: left
    x1 : float
    #: top
    y1 : float
    #: right
    x2 : float
    #: bottom
    y2 : float

    @property
    def w(self):
        """ self width """
        return self.x2 - self.x1

    @property
    def h(self):
        """ self height """
        return self.y2 - self.y1

    @property
    def center(self):
        """ center (x,y) of self (x1, y1, x2, y2) """
        x1, y1, x2, y2 = self
        return ((x1 + x2) / 2), ((y1 + y2) / 2)

    @property
    def area(self):
        """ self Surface area """
        return self.w * self.h

    @property
    def max_dim_len(self):
        """ max (width, height)"""
        return max(self.h, self.w)

    @property
    def square(self):
        """ returns the smallest square containing the rectangle"""
        offset = self.max_dim_len / 2
        xc, yc = self.center
        return Rect(xc - offset, yc - offset, xc + offset, yc + offset)


    def transpose(self, xoffset, yoffset):
        """
        Translation

        Args:
            xoffset (float): horizontal offset.
            yoffset (float): vertical offset.
        Returns:
            Rect :
        """
        x1, y1, x2, y2 = self
        return Rect(x1 + xoffset, y1 + yoffset, x2 + xoffset, y2 + yoffset)

    def mult(self, x, y):
        """
        Multiply self coordinates by horizontal and vertical scaling factors
        Usefull for converting [0...1] coordinates to image frame dimensions in pixels

        Args:
            x (float: horizontal scaling factor.
            y (float): vertical scaling factor.

        Returns:
            Rect:

        """
        x1, y1, x2, y2 = self
        return Rect(x1 * x, y1 * y, x2 * x, y2 * y)

    def intersect(self, r):
        """
        Rectangle intersection between self and r

        Args:
            r (Rect):

        Returns:
            Rect:

        """
        x1, y1, x2, y2 = self
        ret = Rect(max(x1, r.x1), max(y1, r.y1), min(x2, r.x2), min(y2, r.y2))
        if ret.h <= 0 or ret.w <= 0:
            return Rect(0,0,0,0)
        return ret

    def iou(self, r):
        """
        Intersection Over Union between self and r

        Args:
            r (Rect):

        Returns:
            float:

        """
        inter = self.intersect(r).area
        union = self.area + r.area - inter
        return inter / union

    def __contains__(self, point):
        """
        point contained in self

        Args:
            point (tuple): (x,y)

        Returns:
            bool: True if point is in self, else False

        """

        x1, y1, x2, y2 = self
        x, y = point
        return x >= x1 and x <= x2 and y >= y1 and y <= y2

    def to_int(self):
        """
        Convert self coordinates (float) to the nearest int values
        Returns:
            Rect:
        """
        return Rect(*[int(round(e)) for e in self])

    @staticmethod
    def from_dlib(x):
        """
        create Rect from dlib's rectangle instance

        Args:
            x (dlib.rectangle or dlib.drectangle):

        Returns:
            Rect:

        """
        return Rect(x.left(), x.top(), x.right(), x.bottom())


    def to_dlibInt(self):
        """
        Convert self to dlib.rectangle (int)

        Returns:
            dlib.rectangle:

        """
        return dlib.rectangle(*[e for e in self.to_int()])

    def to_dlibFloat(self):
        """
        Convert self to dlib.drectangle (float)

        Returns:
            dlib.drectangle:

        """
        return dlib.drectangle(*self)


    def scale(self, scale_prct):
        """
        scale self according to a given scale percentage

        Args:
            scale_prct (float):

        Returns:
            Rect:

        """
        w, h = (self.w, self.h)
        x1, y1, x2, y2 = self
        xdiff = (w * scale_prct - w) / 2
        ydiff = (h * scale_prct -h) / 2
        return Rect(x1 - xdiff, y1 - ydiff, x2 + xdiff, y2 + ydiff)




#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2019-2021 Ina (David Doukhan & Zohra Rezgui - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module :mod:`inaFaceAnalyzer.display_utils` contains functions allowing to
export video analysis results to formats allowing to display incrusted face
detection bounding boxes and classification estimates.

- :func:`ass_subtitle_export` allows to export results as ASS subtitles (faster).
- :func:`video_export` generate a new video with incrusted information (longer).

Display functions are currenly limited to the results obtained
with :class:`inaFaceAnalyzer.inaFaceAnalyzer.VideoAnalyzer` and
:class:`inaFaceAnalyzer.inaFaceAnalyzer.VideoTracking` analysis pipelines.

>>> from inaFaceAnalyzer.inaFaceAnalyzer import VideoAnalyzer
>>> from inaFaceAnalyzer.display_utils import ass_subtitle_export
>>> va = VideoAnalyzer()
>>> input_vid = './media/pexels-artem-podrez-5725953.mp4'
>>> # define analysis_fps=2 in order to process 2 image frames per second of video
>>> # analysis_fps should be used for analysis AND subtitle export
>>> analysis_fps = 2
>>> df = va(input_vid, fps=analysis_fps)
>>> # export results to ass subtitle
>>> ass_subtitle_export(vid_src, df, './mysubtitle.ass', analysis_fps=analysis_fps)
"""

import cv2
import tempfile
import os
import pandas as pd
from Cheetah.Template import Template
import datetime
from .opencv_utils import video_iterator, get_video_properties, analysisFPS2subsamp_coeff

def _hex2rgb(hx):
    return (int(hx[0:2],16), int(hx[2:4],16), int(hx[4:6],16))


def _sec2hmsms(s):
    td = datetime.timedelta(seconds=s)
    h,m,s = str(td).split(':')
    return '%d:%d:%.2f' % (int(h), int(m), float(s))


def _analysis2displaydf(df, fps, subsamp_coeff, text_pat = None, cols=None):
    """
    Convert analysis results to a generic pandas dataframe containing
    formated information to be displayed using export functions defined bellow
    """
    if isinstance(df, str):
        df = pd.read_csv(df)
        df.bbox = df.bbox.map(lambda x: eval(x))

    ret = pd.DataFrame()
    ret['frame'] = df.frame
    ret['bbox'] = df.bbox
    ret[['x1', 'y1', 'x2', 'y2']] = df.apply(lambda x: x.bbox, axis=1, result_type="expand")
    ret['start'] = df.frame.map(lambda x: _sec2hmsms(x / fps))
    ret['stop'] = df.frame.map(lambda x: _sec2hmsms((x + subsamp_coeff) / fps))
    if text_pat is None:
        if 'face_id' in df.columns:
            ret['rgb_color'] = df.sex_label_avg.map(lambda x: '0000FF' if x == 'm' else '00FF00')
            text_pat = 'id: %s'
            cols = ['face_id']
            if 'sex_label_avg' in df.columns:
                text_pat += ' - sex: %s (%.1f)'
                cols += ['sex_label_avg', 'sex_decfunc_avg']
            if 'age_label_avg' in df.columns:
                text_pat += ' - age: %.1f'
                cols += ['age_label_avg']
        else:
            text_pat = 'sex: %s (%.1f)'
            cols = ['sex_label', 'sex_decfunc']
            ret['rgb_color'] = df.sex_label.map(lambda x: '0000FF' if x == 'm' else '00FF00')
            if 'age_label' in df.columns:
                text_pat += '- age: %.1f'
                cols += ['age_label']

    ret['bgr_color'] = ret.rgb_color.map(lambda x: x[4:] + x[2:4] + x[:2])
    ret['text'] = df.apply(lambda x: text_pat % tuple([x[e] for e in cols]), axis=1)
    return ret

def ass_subtitle_export(vid_src, result_df, ass_dst, analysis_fps=None):
    """
    Export inaFaceAnalyzer results to
    `ASS subtitles <https://en.wikipedia.org/wiki/SubStation_Alpha>`_ .
    ASS can embed complex shapes such as annotated face bounding boxes and
    classification predictions.

    Subtitles are a good option for sharing results, since they do not require
    a large amount of storage size, and do not alter original videos.
    Ass subtitles can be  displayed in `VLC <https://www.videolan.org/vlc/>`_,
    `Aegisub <http://www.aegisub.org/>`_
    or `ELAN <https://archive.mpi.nl/tla/elan>`_ annotation software.

    >>> # displaying mysample_FP2.ass subtitle with vlc
    >>> vlc --sub-file ./mysample_FPS2.ass ./sample_vid.mp4

    Args:
        vid_src (str): path to the input video.
        result_df (str or pandas.DataFrame): video analysis result provided as :class:`pandas.DataFrame` or path to saved csv.
        ass_dst (str): output filepath used to save the resulting subtitle. Must have ass extension.
        analysis_fps (numeric or None, optional): Amount of frames per second which were analyzed \
            (fps analysis argument) \
            if set to None, then consider that all video frames were processed. Defaults to None.
    """


    assert ass_dst[-4:].lower() == '.ass', ass_dst

    video_props = get_video_properties(vid_src)
    fps, width, height = [video_props[e] for e in ['fps', 'width', 'height']]

    if analysis_fps is None:
        subsamp_coeff = 1
    else:
        subsamp_coeff = analysisFPS2subsamp_coeff(vid_src, analysis_fps)

    displaydf = _analysis2displaydf(result_df, fps, subsamp_coeff)

    p = os.path.dirname(__file__)
    t = Template(file = p + '/template.ass')

    t.height = height
    t.width = width
    t.display_df = displaydf
    # text font size set to 4% video height
    t.text_font_size = int(0.04 * height)

    with open(ass_dst, 'wt') as fid:
        print(t, file=fid)


def video_export(vid_src, result_df, vid_dst, analysis_fps=None):
    """
    Export inaFaceAnalyzer results to a video with incrusted faces bounding
    boxes and other analysis information.

    Args:
        vid_src (str): path to the input video.
        result_df (str or pandas.DataFrame): video analysis result provided as \
            :class:`pandas.DataFrame` or path to saved csv.
        vid_dst (str): output path of the resulting video. Must have MP4 extension.
        analysis_fps (int, optional): Amount of frames per second which were analyzed \
            (fps analysis argument). If set to None, then consider that all \
                video frames were processed. Defaults to None.
    """
    assert vid_dst[-4:].lower() == '.mp4', vid_dst


    video_props = get_video_properties(vid_src)
    fps, width, height = [video_props[e] for e in ['fps', 'width', 'height']]

    if analysis_fps is None:
        subsamp_coeff = 1
    else:
        subsamp_coeff = analysisFPS2subsamp_coeff(vid_src, analysis_fps)

    displaydf = _analysis2displaydf(result_df, fps, subsamp_coeff)

    font = cv2.FONT_HERSHEY_SIMPLEX

    with tempfile.TemporaryDirectory() as p:
        tmpout = '%s/%s.mp4' % (p, os.path.splitext(os.path.basename(vid_src))[0])

        fourcc = cv2.VideoWriter_fourcc(*'mp4v')
        out = cv2.VideoWriter(tmpout, fourcc, fps / subsamp_coeff, (width, height))
        #print('out', out)

        for iframe, frame in video_iterator(vid_src, subsamp_coeff=subsamp_coeff):

            sdf = displaydf[displaydf.frame == iframe]

            for e in sdf.itertuples():

                x1, y1, x2, y2 = e.bbox
                color = _hex2rgb(e.rgb_color)
                cv2.putText(frame,e.text,(x1 - 100, y1 - 10 ), font, 0.7, color,2,cv2.LINE_AA)
                cv2.rectangle(frame, (x1, y1), (x2, y2), color, 8)

            ret = out.write(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))

        out.release()
        os.system('ffmpeg -y -i %s -i %s -map 0:v:0 -map 1:a? -vcodec libx264 -acodec copy %s' % (tmpout, vid_src, vid_dst))




#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2019 Ina (David Doukhan - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from .inaFaceAnalyzer import VideoAnalyzer, ImageAnalyzer, VideoTracking
#from .display_utils import incrust_faces_in_video

from . import _version
__version__ = _version.get_versions()['version']



#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2019-2021 Ina (David Doukhan - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import Pyro4
import os
import socket
from collections import namedtuple
import traceback
import time
import inaFaceAnalyzer.commandline_utils as ifacu
from inaFaceAnalyzer.display_utils import ass_subtitle_export, video_export


class IfaWorker:
    def __init__(self, server_uri, batch_size, name_suffix=''):
        # setup network configuration
        self.hostname = hostname = (socket.gethostname() + name_suffix)
        self.jobserver = jobserver = Pyro4.Proxy(server_uri)

        # get processing settings from server, and update with command line arguments
        server_args = jobserver.get_analysis_args('initializing ' + hostname)
        print('recieved args', server_args)
        server_args['batch_size'] = batch_size
        nt = namedtuple('inaFaceAnalyzerArgs', server_args)
        args = nt(**server_args)

        # to be used at engine inference
        self.fps = args.fps

        ## perform engine instantiation
        self.engine = ifacu.engine_factory(args)
        self.msg = 'first call'

    def __call__(self):
        # ask for a job
        # return False if no more jobs are available, else True
        try:
            msg = '%s: %s' % (self.hostname, self.msg)
            src, dst_csv, dst_ass, dst_mp4 = self.jobserver.get_job(msg)
        except StopIteration:
            print('all jobs are done')
            return False

        # init process
        print('received job', src, dst_csv, dst_ass, dst_mp4)
        self.msg = ''
        df = None

        # taskid values: 0 for main analysis, 1 for ass export, 2 for mp4 export
        for taskid, dst in enumerate([dst_csv, dst_ass, dst_mp4]):

            # test if destination path is provided: not None and not Nan
            if not (dst and (dst == dst)):
                continue

            # skip if file already exists
            if os.path.exists(dst):
                self.msg += '%s already exists' % dst
                continue

            # do the job
            b = time.time()
            try:
                # create output directory if needed
                os.makedirs(os.path.dirname(dst), exist_ok=True)

                # task specific action
                if taskid == 0:
                    df = self.engine(src, fps = self.fps)
                    df.to_csv(dst, index=False, float_format='%.2f')
                else:
                    if df is None:
                        df = dst_csv
                    if taskid == 1:
                        print(src, dst, self.fps, df)
                        ass_subtitle_export(src, df, dst, analysis_fps = self.fps)
                    elif taskid == 2:
                        video_export(src, df, dst, analysis_fps = self.fps)
                    else:
                        raise NotImplementedError()

                self.msg += '%s done in %.1f sec' % (dst, time.time() - b)
            except BaseException:
                self.msg += 'problem with %s\n' % dst
                self.msg += traceback.format_exc()

        return True


description = '''Worker in charge of analyzing documents and receive orders
from a central job server. Workers need to have access to input files (url or path)
and must be allowed to write results in a centralized storage directory
To be used jointly with ina_face_analyzer_distributed_server.py .
The server must be run before runing the worker.
'''

hserver_uri = '''
The URI displayed by the server to be copy/pasted after runing program
ina_face_analyzer_distributed_server.py  .
Sample URI: PYRO:obj_25e3ee9d312848fcaf0784c2b80933c4@blahtop:44175
'''


if __name__ == '__main__':
    # parse command line arguments
    parser = ifacu.new_parser(description)
    parser.add_argument(dest='server_uri', help=hserver_uri)
    ifacu.add_batchsize(parser)
    args = parser.parse_args()

    # create worker instance
    worker = IfaWorker(args.server_uri, args.batch_size)

    # iterate while jobs are available
    while worker():
        pass



#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2019-2021 Ina (David Doukhan - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.# -*- coding: utf-8 -*-

import Pyro4
import pandas as pd
import inaFaceAnalyzer.commandline_utils as ifacu
from inaFaceAnalyzer.face_detector import facedetection_cmdline
from inaFaceAnalyzer.face_classifier import faceclassifier_cmdline

import warnings

@Pyro4.expose
class JobServer(object):
    def __init__(self, args):
        '''
        Create an instance of JobServer, in charge of distributing analyses to
        be performed to a set of remote workers.

        Parameters
        ----------
        args : argument parser results
            see ina_face_analyzer_distributed_server.py -h for a detailed listing
        '''


        # open csv and check mandatory columns are provided
        csv = args.joblist_csv
        print('parsing joblist %s' % csv)
        df = pd.read_csv(csv)
        # check mandatory columns are provided
        assert 'source_path' in df.columns, '%s must contain a source_path column' % csv
        assert 'dest_csv' in df.columns, '%s must contain a dest_csv column' % csv
        for k in df.columns:
            if k not in ['source_path', 'dest_csv', 'dest_ass', 'dest_mp4']:
                warnings.warn('column %s in %s is not supported and will be ignored' % (k, csv))
            df[k] = df[k].map(lambda x: x.strip() if isinstance(x, str) else x)
        # shuffle records and drop duplicates
        df = df.drop_duplicates().sample(frac=1).reset_index(drop=True)
        if 'dest_ass' not in df:
            df['dest_ass'] = None
        if 'dest_mp4' not in df:
            df['dest_mp4'] = None


        print('setting jobs')
        print('random job record example:', next(df.itertuples()))
        self.nbjobs = len(df)
        print('Total number of files to process:', self.nbjobs)
        self.jobiterator = enumerate(df.itertuples())
        self.args = args.__dict__

    def get_analysis_args(self, msg):
        print(msg)
        return self.args

    def get_job(self, msg):
        print('receiving message: ' + msg)
        ijob, job = next(self.jobiterator)
        ret = (job.source_path, job.dest_csv, job.dest_ass, job.dest_mp4)
        print ('sending job %d/%d: %s' % (ijob+1, self.nbjobs, ret))
        return ret

    # to be implemented - only usefull for image collections
    # def get_njobs(self, msg, nbjobs=20):
    #     print('jobs %d-%d: %s' % (self.i, self.i + nbjobs, msg))
    #     ret = (self.lsource[:nbjobs], self.ldest[:nbjobs])
    #     if len(ret[0]) == 0:
    #         print('All jobs dispatched')
    #     self.lsource = self.lsource[nbjobs:]
    #     self.ldest = self.ldest[nbjobs:]
    #     self.i += nbjobs
    #     return ret

description = '''Server in charge of distributing a list of documents to analyze to
workers distributed over the network. Workers need to have access to a
centralized storage directory for writing output results (mounted with
NFS, sshfs, ...).
To be used jointly with ina_face_analyzer_distributed_worker.py.'''

hhostaddress = '''host_address is used by workers to communicate remotely with
the server. It can be either server's IP adress, or host full name ex: mymachine.my-domain.fr
'''

hjoblist = ''' joblist_csv is the full path to a csv file storing one line per file
to process. It may contain up to 4 columns, separated by coma ",".
"source_path" column (mandatory) contains the path or the url to a file to be processed.
"dest_csv" (mandatory) contains the path used to write the resulting csv.
"dest_ass" (non mantatory) is a path used to export results to ASS rich subtitles (displayed in VLC or ELAN).
"dest_mp4" (non mandatory) is a path used to export results to MP4 video
with incrusted face bouding boxes and classification results.
'''

hengine = '''Analysis engine to be used.
"video" should be the default choice for video materials.
"videotracking" use face tracking methods to lower computation time and smooth
face classification results obtained on video materials.
Tracked faces are associated to a numeric identifier.
Smoothed classification estimates are associated to "_avg" suffix in resulting column names,
and are more robust than frame-isolated predictions.
"videokeyframes" restrict analyses to video key frames. It allows the fastest
video analysis sumary, but is is associated to a non uniform frame sampling rate.
'''

if __name__ == '__main__':

    parser = ifacu.new_parser(description)

    parser.add_argument(dest='host_address', help = hhostaddress)
    parser.add_argument(help = hjoblist, dest='joblist_csv')


    parser.add_argument('--engine', choices=['video', 'videotracking', 'videokeyframes'],
                required=True, help=hengine)

    faceclassifier_cmdline(parser)
    ifacu.add_fps(parser)
    ifacu.add_tracking(parser)
    facedetection_cmdline(parser)

    #### OPTION SKIP IF EXIST

    # parse command line arguments
    args = parser.parse_args()

    # full name of the host to be used by remote clients
    Pyro4.config.HOST = args.host_address
    daemon = Pyro4.Daemon()

    uri = daemon.register(JobServer(args))
    print("Provide the following objet URI to remote ina_face_analyzer_distributed_workers: ", uri)
    daemon.requestLoop()



#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2021 Ina (David Doukhan - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
from tensorflow.keras.utils import get_file

# inaFaceAnalyzer models are stored remotely within github releases
# this code allows to download and use them on demand

r1_url = 'https://github.com/ina-foss/inaFaceAnalyzer/releases/download/models/'


# TODO: define RC MALLI MODEL ALSO HERE
dmodels = {
    # These 2 models are provided for face detection in opencv
    # https://raw.githubusercontent.com/opencv/opencv_3rdparty/dnn_samples_face_detector_20180220_uint8/opencv_face_detector_uint8.pb
    # https://github.com/opencv/opencv/blob/4eb296655958e9241f43a7bd144d5e63759f6cea/samples/dnn/face_detector/opencv_face_detector.pbtxt
    'opencv_face_detector_uint8.pb' : r1_url,
    'opencv_face_detector.pbtxt' : r1_url,
    # Dlib's Facial landmarks prediction
    # http://dlib.net/files/shape_predictor_68_face_landmarks.dat.bz2
    'shape_predictor_68_face_landmarks.dat' : r1_url,
    # Resnet50 architectures trained on FairFace database
    'keras_resnet50_fairface_GRA.h5' : r1_url,
    'keras_resnet50_fairface.h5' : r1_url,
    # linear SVM trained on Youtube Faces VGG16 face embeddings
    'svm_ytf_zrezgui.hdf5' : r1_url,
    # linear SVM trained on FairFace VGG 16 embeddings
    'svm_vgg16_fairface.hdf5' : r1_url,
    # face detection and landmark estimation proovided in libfacedetection
    # https://github.com/ShiqiYu/libfacedetection
    'libfacedetection-yunet.onnx' : r1_url}

def get_remote(model_fname):
    # if in a docker image, try to get the file in /root/.keras
    rootpath = '/root/.keras/inaFaceAnalyzer/' + model_fname
    if os.access(rootpath, os.R_OK):
        return rootpath

    # standard keras get file
    # check if the file is in /home/$USER/.keras/inaFaceAnalyzer/ and download it if required
    url = dmodels[model_fname]
    return get_file(model_fname, url + model_fname, cache_subdir='inaFaceAnalyzer')




#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2019-2022 Ina (David Doukhan - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#import argparse
import os
import sys

import inaFaceAnalyzer.commandline_utils as ifacu
import inaFaceAnalyzer.face_classifier
from inaFaceAnalyzer.face_detector import facedetection_cmdline
from inaFaceAnalyzer.face_classifier import faceclassifier_cmdline

import inaFaceAnalyzer.display_utils


# TODO: add options to perform detection only and/or preprocessing only
# options to export image results to PDF with incrusted bounding boxes ?


description = 'inaFaceAnalyzer %s: detects and classify faces from media collections and export results in csv' % inaFaceAnalyzer.__version__
parser = ifacu.new_parser(description)

## Required arguments
ra = parser.add_argument_group('required arguments')

hengine = '''Analysis engine to be used.
"video" should be the default choice for video materials.
"image" should be the default choice for image files.
"videotracking" uses face tracking methods to lower computation time and smooth
face classification results obtained on video materials.
Tracked faces are associated to a numeric identifier.
Smoothed classification estimates are associated to "_avg" suffix in resulting column names,
and are more robust than frame-isolated predictions.
"videokeyframes" restricts analyses to video key frames. It allows the fastest
video analysis summary, but is is associated to a non-uniform frame sampling rate.
"preprocessed_image" skips face detection and alignment steps and requires face
images that can be used directly by face classifiers.
These images should be already detected, cropped, centered, aligned and rescaled to 224*224 pixels.
'''
ra.add_argument('--engine', choices=['video', 'image', 'videotracking', 'videokeyframes', 'preprocessed_image'],
                required=True, help=hengine)

h = '''INPUT is a list of documents to analyse. ex: /home/david/test.mp4 /tmp/mymedia.avi.
INPUT can be a list of video paths OR a list of image paths.
Videos and images can have heterogenous formats but cannot be mixed in a single command.
'''
ra.add_argument('-i', nargs='+', required=True, dest='input', help = h)

h = '''When used with "video" and "videotracking" engines, OUTPUT is the path
to an existing directory storing one resulting CSV for each processed video.
When used with "image" and "preprocessed_image" engines, OUTPUT is the path to
a resulting csv that will be created and will contain a line for each detected faces.
'''
ra.add_argument('-o', required=True, help = h, dest='output')


## Optional arguments
oa = parser.add_argument_group('optional arguments')

#oa.add_argument("-h", "--help", action="help", help="show this help message and exit")


# classifier
faceclassifier_cmdline(parser)

ifacu.add_batchsize(oa)

## Video only parameters
dv = parser.add_argument_group('optional arguments to be used only with "video" and "videotracking" engines')
ifacu.add_fps(dv)

dv.add_argument('--ass_subtitle_export', action='store_true',
                help='export analyses into a rich ASS subtitle file which can be displayed with VLC or ELAN')

dv.add_argument('--mp4_export', action='store_true',
                help='export analyses into a a MP4 video with incrusted bounding boxes and analysis estimates')


ifacu.add_tracking(parser)

# face detection
facedetection_cmdline(parser)


#### parse arguments
args = parser.parse_args()


# deal with incompatible arguments

class ParserError:
    def __init__(self, parser):
        self.parser = parser
    def __call__(self, a1, a2):
        raise self.parser.error('%s argument cannot be mixed with %s' % (a1, a2))
pe = ParserError(parser)

# Management of incompatible uses

for k in ['fps', 'ass_subtitle_export', 'mp4_export']:
    if args.__dict__[k] and (args.engine not in ['video', 'videotracking']):
        pe('--' + k, '--engine ' + args.engine)


# analysis engine constructor
engine = ifacu.engine_factory(args)

# image engines
if args.engine in ['image', 'preprocessed_image']:
    # check provided output extension
    out_ext = os.path.splitext(args.output)[1]
    if out_ext.lower() != '.csv':
        raise ValueError('OUTPUT value %s provided to -o argument should have .csv extension' % args.output)

    # run engine
    if args.engine == 'image':
        df = engine(args.input)
    elif args.engine == 'preprocessed_image':
        df = engine.preprocessed_img_list(args.input, batch_len=args.batch_size)
    # save results
    df.to_csv(args.output, index=False)
    sys.exit(0)



# VIDEO engines

# test that provided output path is an existing directory
if not os.path.isdir(args.output):
    raise ValueError('OUTPUT directory %s provided to -o argument should be an existing directory' % args.output)

dargs = {}
if args.fps:
    dargs = {'fps': args.fps}


nbvid = len(args.input)
for i, f in enumerate(args.input):
    # TODO: add a try/catch system
    # TODO: add a progressbar.ProgressBar
    print('analyzing video %d/%d: %s' % (i, nbvid, f))
    base, _ = os.path.splitext(os.path.basename(f))
    df = engine(f, **dargs)
    df.to_csv('%s/%s.csv' % (args.output, base), index=False)
    if args.ass_subtitle_export:
        inaFaceAnalyzer.display_utils.ass_subtitle_export(f, df, '%s/%s.ass' % (args.output, base), analysis_fps=args.fps)
    if args.mp4_export:
        inaFaceAnalyzer.display_utils.video_export(f, df, '%s/%s.mp4' % (args.output, base), analysis_fps=args.fps)

